@@ -22,14 +22,7 @@ import os, struct, stat, os.path, re, commands
 from mx import DateTime
 
 from cfvers.main import *
-from cfvers.cmd  import *
 
-version_info = (0, 4, 3, 'alpha', 0)
-hexversion   = 0x00040200
-ident_rev = """$Id: __init__.py 67 2003-10-26 20:04:45Z iusty $"""
-
-class CfversException(Exception):
-    pass
-
-class ConfigException(CfversException):
-    pass
+version_info = (0, 4, 6, 'alpha', 0)
+hexversion   = 0x00040600
+ident_rev = """$Id: __init__.py 77 2003-11-09 22:48:46Z iusty $"""
