@@ -21,21 +21,20 @@
 import os, struct, stat, os.path, re, commands, sys
 import types
 import random
+import errno
 from mx import DateTime
 from ConfigParser import SafeConfigParser
+import tarfile
+from cStringIO import StringIO
 
 import cfvers
 import cfvers.repository
-from main import *
+import cfvers.main
+from cfvers.main import *
 
-#Return codes for store operation
-STORED_OK = 0
-STORED_NOTCHANGED = 1
-STORED_IOERROR = 2
-
-__all__ = ["STORED_OK", "STORED_NOTCHANGED", "STORED_IOERROR",
-           "AdminCommands", "Commands",
-           ]
+__all__ = [
+    "AdminCommands", "Commands",
+    ]
 
 class CLIScript(object):
     """Base class for command line handlers"""
@@ -86,7 +85,7 @@ FOR A PARTICULAR PURPOSE.""" % nv
         if revstring is None:
             return None, None
         # match and extract from glob rev[:rev]
-        m=re.match("^(?P<start>[0-9]+)(:(?P<end>[0-9]+))?$", revstring)
+        m=re.match("^((?P<start>[0-9]+):)?(?P<end>[0-9]+)$", revstring)
         if m is None:
             raise ValueError, "Invalid syntax in revision specification %s" % revstring
         r1 = m.group('start')
@@ -143,7 +142,7 @@ class Commands(CLIScript):
         self.repo.close()
         return
     
-    def _map_files(self, area, files=None, norecurse=0):
+    def _map_files(self, area, files=None, norecurse=False):
         """Maps a (possibly empty) list of files to a list of targets"""
         
         def helper(arg, dirname, names):
@@ -157,38 +156,36 @@ class Commands(CLIScript):
                 realdir = dirname[len(path):]
             else:
                 raise ValueError, "Path doesn't lie in its area!"
-            #print path, realdir
             mylist.extend([os.path.join(realdir, fname) for fname in names])
             return
 
         repo = self.repo
         if len(files) == 0:
-            subfiles = []
-            for item in repo.items():
-                latest = repo.getEntry(item, None)
+            for item in repo.items(area=area):
+                latest = repo.getEntry(item, None, do_payload=False)
                 if latest is not None:
-                    subfiles.append(latest.filename)
+                    files.append(latest.filename)
                 else:
-                    subfiles.append(item.name)
-            files = subfiles
+                    files.append(item.name)
             
         files = map(os.path.abspath, files)
+        files = dict.fromkeys(files).keys()
         targets = []
         errors = []
-        for f in files:
-            targets.append(f)
-            realf = forcejoin(area.root, f)
-            if not norecurse:
-                try:
-                    isdir = stat.S_ISDIR(os.lstat(realf).st_mode)
-                except IOError, e:
-                    errors.append((STORED_IOERROR, e, None, None))
-                    continue
-                except OSError, e:
-                    errors.append((STORED_IOERROR, e, None, None))
-                    continue
-                if isdir:
-                    os.path.walk(realf, helper, (targets, area.root))
+        for virtf in files:
+            realf = forcejoin(area.root, virtf)
+            try:
+                st = os.lstat(realf)
+            except EnvironmentError, e:
+                errors.append(Result(Result.STORED_IOERROR, exc=e))
+                continue
+            # If we got here, the file at least exists,
+            # so we'll mark to process
+            targets.append(virtf)
+            if not norecurse and stat.S_ISDIR(st.st_mode):
+                os.path.walk(realf, helper, (targets, area.root))
+        targets = dict.fromkeys(targets).keys()
+        targets.sort()
         return (targets, errors)
         
     def store(self, files=[], options=None):
@@ -207,22 +204,21 @@ class Commands(CLIScript):
         repo = self.repo
         a = self.area
         (targets, errors) = self._map_files(a, files=files, norecurse=norecurse)
-        ar = AreaRev(a, logmsg)
+        ar = AreaRev(area=a, logmsg=logmsg, commiter=options.commiter)
         newrev = ar.revno
         stored = 0
         for name in targets:
-            if not options.quiet:
-                print "Storing %s..." % name
             item = repo.getItemByName(a, name)
             if item is None:
-                item = Item(area=a, name=name)
+                item = Item(area=a, name=name,
+                            metadata_only=options.nopayload)
                 repo.addItem(item)
                 item = repo.getItemByName(a, name)
+                item.metadata_only = options.nopayload
             retcode = self._store_newrev(item, newrev)
-            if retcode[0] != STORED_OK:
-                errors.append(retcode)
-            else:
+            if retcode.code == Result.STORED_OK:
                 stored += 1
+            errors.append(retcode)
 
         if stored > 0:
             repo.putAreaRev(ar)
@@ -231,22 +227,20 @@ class Commands(CLIScript):
         else:
             repo.rollback()
             store_done = False
-        return errors, store_done
+        return errors, store_done, newrev
 
     def _store_newrev(self, item, revno):
         try:
             rev = RevEntry(item, revno)
-        except IOError, e:
-            return (STORED_IOERROR, e, item, None)
-        except OSError, e:
-            return (STORED_IOERROR, e, item, None)
+        except EnvironmentError, e:
+            return Result(Result.STORED_IOERROR, item=item, exc=e)
         
         old = self.repo.getEntry(item, None)
         if (not old is None) and rev == old:
-            return (STORED_NOTCHANGED, 'Item has not changed, not stored', item, old)
+            return Result(Result.STORED_NOTCHANGED, item=item, rev=old)
         else:
             self.repo.addEntry(rev)
-        return (STORED_OK, '', item, rev)
+        return Result(Result.STORED_OK, item=item, rev=rev)
 
     def diff(self, options=None, files=None):
         rev1, rev2 = self.parserev(options.rev)
@@ -263,6 +257,14 @@ class Commands(CLIScript):
         return
                 
     def _diff_file(self, vi, rev1, rev2, options):
+        """Execute a diff between two revisions.
+
+        Parameters:
+          - vi: the item which to analyze
+          - rev1: the source revision
+          - rev2: the target revision
+
+        """
         rlist = self.repo.getRevList(vi)
         if len(rlist) == 0:
             if options.list:
@@ -270,6 +272,11 @@ class Commands(CLIScript):
             else:
                 print "===== Item %s does not have any revisions" % vi.name
             return
+        # rev1 is older (source)
+        # rev2 is newer (target)
+        if rev1 is None:
+            rev1 = rev2
+            rev2 = None
         e1 = self.repo.getEntry(vi, rev1)
         if e1 is None:
             if options.list:
@@ -317,21 +324,22 @@ class Commands(CLIScript):
         return
 
     def retrieve(self, files=None, options=None):
+        results = []
         if len(files) != 0:
             mlist = []
             for filename in files:
                 af = os.path.abspath(filename)
                 item = self.repo.getItemByName(self.area, af)
                 if item is None:
-                    print "Skipped: file '%s' is not being tracked!" % af
+                    results.append(Result(Result.RETR_NTRACK, fname=af))
                     continue
                 mlist.append(item)
                 if not options.norecurse:
                     mlist += self._recur_build_list(mlist, [])
         else:
             mlist = self.repo.items(area=self.area)
-        self._retrieve_list(mlist, options)
-        return
+        self._retrieve_list(mlist, options, results)
+        return results
 
     def _recur_build_list(self, litems, alsoskip):
         nlist = []
@@ -346,22 +354,20 @@ class Commands(CLIScript):
         return nlist
 
     def _retrieve_file(self, vi, rev, options=None):
-        if not options.quiet:
-            print "Retrieving: '%s'" % vi.name
-        rev.to_filesys(destdir=options.destdir, use_dirs=options.use_dirs)
-        return
+        return rev.to_filesys(destdir=options.destdir, use_dirs=options.use_dirs)
 
-    def _retrieve_list(self, ilist, options):
+    def _retrieve_list(self, ilist, options, results):
         for item in ilist:
             rev = self.repo.getEntry(item, options.revno)
             if rev is None:
                 revlist = self.repo.getRevList(item)
                 if len(revlist) == 0:
-                    print >>sys.stderr, "Skipped: Item '%s' doesn't have any revisions!" % item.name
+                    ncod = Result.RETR_NOREVS
                 else:
-                    print >>sys.stderr, "Skipped: Item '%s', can't find revision %d!" % (item.name, options.revno)
+                    ncod = Result.RETR_NOXREV
+                results.append(Result(ncod, item=item))
             else:
-                self._retrieve_file(item, rev, options=options)
+                results.append(self._retrieve_file(item, rev, options=options))
         return
 
     def show(self, filename, rev=None):
@@ -377,6 +383,9 @@ class Commands(CLIScript):
             else:
                 print >> sys.stderr, "Can't find revision '%s' for item '%s'!" % (rev, vi.name)
             return
+        if entry.filecontents is None:
+            print >> sys.stderr, "File contents has not been stored for file '%s'!" % entry.filename
+            return
         print entry.filecontents,
         return
 
@@ -397,13 +406,13 @@ class Commands(CLIScript):
                     continue
                 mlist.append(item)
         else:
-            mlist = self.repo.items()
+            mlist = self.repo.items(area=self.area)
         self._stat_list(mlist, version)
         return
 
     def _stat_list(self, ilist, version):
         for item in ilist:
-            rev = self.repo.getEntry(item, None)
+            rev = self.repo.getEntry(item, None, do_payload=False)
             if rev is None:
                 revlist = self.repo.getRevList(item)
                 if len(revlist) == 0:
@@ -422,3 +431,71 @@ class Commands(CLIScript):
     def log(self):
         arearevs = self.repo.getAreaRevs(self.area)
         return arearevs
+
+    def exportcksum(self, cmdopts, cmdargs):
+        if cmdopts.output == "-":
+            f = sys.stdout
+            close = False
+        else:
+            f = file(cmdopts.output, "w")
+            close = True
+        for rev in self.repo.getEntries(self.area, cmdopts.r2, do_payload=False):
+            if rev is not None and cmdopts.r1 is not None and rev.revno < cmdopts.r1:
+                continue
+            if rev is not None and rev.isreg():
+                f.write("%s  %s\n" % (rev.sha1sum, rev.filename[1:]))
+        if close:
+            f.close()
+        return
+
+    def exporttar(self, cmdopts, cmdargs):
+        if cmdopts.output == "-":
+            tarh = tarfile.open(fileobj=sys.stdout, mode="w|")
+        else:
+            tarh = tarfile.open(name=cmdopts.output, mode="w")
+        for rev in self.repo.getEntries(self.area, None):
+            if rev is not None:
+                if rev.isreg():
+                    fdata = StringIO(rev.filecontents)
+                else:
+                    fdata = None
+                tarh.addfile(self._genfakefile(rev), fdata)
+        tarh.close()
+        return
+
+    def _genfakefile(self, rev):
+        """Generate a fake TarInfo object from a revision entry.
+
+        Parameters:
+        rev - the revision entry from which to create the archive member.
+        
+        """
+        ti = tarfile.TarInfo()
+        if rev.filename.startswith("/"):
+            ti.name = rev.filename[1:]
+        else:
+            ti.name = rev.filename
+        ti.uname = ""
+        ti.gname = ""
+        ti.uid = rev.uid
+        ti.gid = rev.gid
+        ti.mtime = rev.mtime
+        ti.mode = rev.mode
+        if rev.isreg():
+            ti.chksum = tarfile.calc_chksum(rev.filecontents)
+            ti.size = rev.size
+            ti.type = tarfile.REGTYPE
+        elif rev.isdir():
+            ti.type = tarfile.DIRTYPE
+        elif rev.islnk():
+            ti.type = tarfile.SYMTYPE
+            ti.linkname = rev.filecontents
+        elif rev.isblk():
+            ti.type = tarfile.BLKTYPE
+        elif rev.ischr():
+            ti.type = tarfile.CHRTYPE
+        elif rev.ififo():
+            ti.type = tarfile.FIFOTYPE
+        elif rev.isock():
+            return None
+        return ti
