@@ -36,7 +36,11 @@ import string
 from cStringIO import StringIO
 from mx import DateTime
 
-__all__ = ["Area", "Item", "RevEntry", "AreaRev", "forcejoin", "rexists"]
+import cfvers
+
+__all__ = ["Area", "Item", "RevEntry", "AreaRev", "forcejoin", "rexists",
+           "CfversException", "ConfigException", "RepositoryException",
+           "Result"]
 
 def forcejoin(a, *p):
     """Join two or more pathname components, considering them all relative"""
@@ -97,15 +101,16 @@ class Area(object):
     
 
 class Item(object):
-    __slots__ = ["id", "area", "name", "ctime", "dirname"]
+    __slots__ = ["id", "area", "name", "ctime", "dirname", "metadata_only"]
     
     def __init__(self, id=-1, area=None, name=None,
-                 ctime=None, dirname=None):
+                 ctime=None, dirname=None, metadata_only=False):
         if name is None or not name.startswith("/"):
             raise ValueError, "Invalid name '%s'" % name
         self.id = id
         self.area = area
         self.name = name
+        self.metadata_only = metadata_only
         if dirname is None:
             self.dirname = os.path.dirname(name)
         else:
@@ -167,6 +172,8 @@ class RevEntry(object):
             self.filecontents = ""
         psha = sha.new(self.filecontents)
         self.sha1sum = psha.hexdigest()
+        if vi.metadata_only and self.filetype == stat.S_IFREG:
+            self.filecontents = None
         self.inode = st.st_ino
         self.device = st.st_dev
         self.nlink = st.st_nlink
@@ -210,6 +217,10 @@ class RevEntry(object):
 
         """
 
+        if self.filecontents is None:
+            return Result(Result.RETR_NODATA, item=self.item, rev=self)
+        retval = Result.RETR_ALLOK
+        retexc = None
         self._check_sum()
         if destdir is None:
             target = self.filename
@@ -221,14 +232,11 @@ class RevEntry(object):
 
         if self.filetype not in (stat.S_IFREG, stat.S_IFLNK, stat.S_IFCHR,
                                  stat.S_IFBLK, stat.S_IFIFO, stat.S_IFDIR):
-            print >>sys.stderr, "Skipping: file type %s (%s) cannot be restored." \
-                  % (self.modemap[self.filetype], self.filename)
-            return
+            return Result(Result.RETR_NA, item=self.item, rev=self)
 
         if (self.filetype == stat.S_IFBLK or self.filetype == stat.S_IFCHR) \
            and self.rdev is None:
-            print >>sys.stderr, "Skipping: st_rdev information not available, can't restore device file!"
-            return
+            return Result(Result.RETR_RDEV, item=self.item, rev=self)
 
         do_create  = True
         do_payload = True
@@ -241,12 +249,11 @@ class RevEntry(object):
             try:
                 os.makedirs(targetbase)
             except OSError, e:
-                print >>sys.stderr, "Warning: While making intermediate directories: %s" % e
+                return Result(Result.RETR_PATHERR, item=self.item, rev=self, exc=e)
 
         if self.filetype == stat.S_IFDIR and rexists(target):
             if not os.path.islink(target) and not os.path.isdir(target):
-                print >>sys.stderr, "Error: Can't overwrite non-directory with directory!"
-                return
+                return Result(Result.RETR_INVTRANS, item=self.item, rev=self)
             else: # The directory already exists; we must restore ownership and attrs
                 do_create  = False
                 do_payload = False
@@ -254,16 +261,14 @@ class RevEntry(object):
                 newname = target
         else:
             if not os.path.islink(target) and os.path.isdir(target):
-                print >>sys.stderr, "Can't overwrite directory with non-directory!"
-                return
+                return Result(Result.RETR_INVTRANS, item=self.item, rev=self)
             newname = "%s.%07d" % (target, random.randint(0, 999999))
             retries = 0
             while rexists(newname) and retries < 1000:
                 newname = "%s.%07d" % (target, random.randint(0, 999999))
                 retries += 1
             if rexists(newname):
-                print >>sys.stderr, "Error: Can't create a temporary filename! Programmer error or race attack?"
-                return
+                return Result(Result.RETR_TEMPFILE, item=self.item, rev=self)
 
         oldumask = os.umask(0777)
         # try...finally for umask restoration
@@ -292,10 +297,8 @@ class RevEntry(object):
                         os.mkdir(newname, 0)
                     else:
                         raise TypeError, "Programming error: shouldn't have to handle file type %s!" % self.modemap[self.filetype]
-                        return
-                except OSError, e:
-                    print >>sys.stderr, "Error: error '%s' while creating temporary file." % e
-                    return
+                except EnvironmentError, e:
+                    return Result(Result.RETR_IOERROR, item=self.item, rev=self, exc=e, descr="while creating file")
                 else:
                     must_remove_new = True
             # From now on, we must cleanup on exit (done via ...finally)
@@ -305,15 +308,16 @@ class RevEntry(object):
                     if self.filetype == stat.S_IFREG:
                         os.write(fd, self.filecontents)
                         os.close(fd)
-                except OSError, e:
-                    print >>sys.stderr, "Error: error '%s' while writing file contents." % e
-                    return
+                except EnvironmentError, e:
+                    return Result(Result.RETR_IOERROR, item=self.item, rev=self, exc=e, descr="while writing file")
             # Step 3
             if do_attrs:
                 try:
                     os.lchown(newname, self.uid, self.gid)
                 except OSError, e:
-                    print >>sys.stderr, "Warning: error '%s' while modifying temporary file ownership." % e
+                    retval = Result.RETR_PARTOK
+                    retexc = e
+                    #print >>sys.stderr, "Warning: error '%s' while modifying temporary file ownership." % e
 
                 # WARNING: don't chmod for symlinks, it acts on the target!!!
                 if not self.filetype == stat.S_IFLNK:
@@ -321,16 +325,14 @@ class RevEntry(object):
                     try:
                         os.chmod(newname, self.mode)
                         os.utime(newname, (self.atime, self.mtime))
-                    except OSError, e:
-                        print >>sys.stderr, "Error: error '%s' while modifying temporary file attributes." % e
-                        return
+                    except EnvironmentError, e:
+                        return Result(Result.RETR_IOERROR, item=self.item, rev=self, exc=e, descr="while modifying attributes")
             # Step 5
             if do_rename:
                 try:
                     os.rename(newname, target)
                 except OSError, e:
-                    print >>sys.stderr, "Error: error '%s' while renaming" % e
-                    raise
+                    return Result(Result.RETR_IOERROR, item=self.item, rev=self, exc=e, descr="while renaming")
                 else:
                     must_remove_new = False # We managed to finish!
         finally:
@@ -340,7 +342,7 @@ class RevEntry(object):
                     os.unlink(newname)
                 except OSError, e:
                     print >>sys.stderr, "Error: while cleaning-up: %s" % e
-        return
+        return Result(retval, item=self.item, rev=self, exc=retexc)
         
     def diff(self, older, options=None):
         obuff = StringIO()
@@ -443,18 +445,30 @@ class RevEntry(object):
         tchar += mapbit(self.mode, stat.S_IRUSR, 'r')
         tchar += mapbit(self.mode, stat.S_IWUSR, 'w')
         if self.mode & stat.S_ISUID:
-            tchar += 'S'
+            if self.mode & stat.S_IXUSR:
+                tchar += 's'
+            else:
+                tchar += 'S'
         else:
             tchar += mapbit(self.mode, stat.S_IXUSR, 'x')
         tchar += mapbit(self.mode, stat.S_IRGRP, 'r')
         tchar += mapbit(self.mode, stat.S_IWGRP, 'w')
         if self.mode & stat.S_ISGID:
-            tchar += 'S'
+            if self.mode & stat.S_IXGRP:
+                tchar += 's'
+            else:
+                tchar += 'S'
         else:
             tchar += mapbit(self.mode, stat.S_IXGRP, 'x')
         tchar += mapbit(self.mode, stat.S_IROTH, 'r')
         tchar += mapbit(self.mode, stat.S_IWOTH, 'w')
-        tchar += mapbit(self.mode, stat.S_IXOTH, 'x')
+        if self.mode & stat.S_ISVTX:
+            if self.mode & stat.S_IXOTH:
+                tchar += 't'
+            else:
+                tchar += 'T'
+        else:
+            tchar += mapbit(self.mode, stat.S_IXOTH, 'x')
         return tchar
 
     def _check_sum(self):
@@ -534,3 +548,94 @@ class AreaRev(object):
 
     ctime = property(_get_ctime, _set_ctime, None,
                      "The creation time of this revision")
+
+class CfversException(Exception):
+    pass
+
+class ConfigException(CfversException):
+    pass
+
+class RepositoryException(CfversException):
+    pass
+
+class Result(object):
+    # Codes for retrieve operation
+    RETR_ALLOK = 0
+    RETR_PARTOK = 1
+    RETR_NTRACK = 2
+    RETR_IOERROR = 3
+    RETR_NA = 4
+    RETR_RDEV = 5
+    RETR_PATHERR = 6
+    RETR_INVTRANS = 7
+    RETR_TEMPFILE = 8
+    RETR_NOREVS = 9
+    RETR_NOXREV = 10
+    RETR_NODATA = 11
+    
+    #Return codes for store operation
+    STORED_OK = 1001
+    STORED_NOTCHANGED = 1002
+    STORED_IOERROR = 1003
+    
+    _codeinfo = {
+        RETR_ALLOK: "Retrieved (fully)",
+        RETR_PARTOK: "Retrieved (partially)",
+        RETR_NTRACK: "Skipped (not versioned)",
+        RETR_IOERROR: "Skipped (error)",
+        RETR_NA: "Skipped (non-retrievable)",
+        RETR_RDEV: "Skipped (rdev info missing)",
+        RETR_PATHERR: "Skipped (path error)",
+        RETR_INVTRANS: "Skipped (invalid transition)",
+        RETR_TEMPFILE: "Skipped (can't create tempfile)",
+        RETR_NOREVS: "Skipped (no revisions)",
+        RETR_NOXREV: "Skipped (no such revision)",
+        RETR_NODATA: "Skipped (only metadata versioned)",
+        
+        STORED_OK: "Stored",
+        STORED_NOTCHANGED: "Skipped (not changed)",
+        STORED_IOERROR: "Skipped (error)",
+        }
+
+    def __init__(self, code, item=None, rev=None,
+                 exc=None, fname=None, descr=None):
+        self.code = code
+        self.item = item
+        self.rev = rev
+        self.exc = exc
+        self.descr = descr
+        self.critical = code not in (self.RETR_ALLOK,
+                                     self.RETR_NA,
+                                     self.STORED_OK,
+                                     self.STORED_NOTCHANGED,
+                                     )
+        # Deduct error information (if applicable)
+        if exc is None:
+            errinfo = ""
+        elif hasattr(exc, "strerror") and hasattr(exc, "errno"):
+            errinfo = "%s (%s)" % (exc.strerror, errno.errorcode[exc.errno])
+        else:
+            errinfo = str(exc)
+        self.errinfo = errinfo
+        
+        # Deduct file name information
+        if fname is not None:
+            self.fname = fname
+        elif hasattr(exc, "filename"):
+            self.fname = exc.filename
+        elif hasattr(rev, "filename"):
+            self.fname = rev.filename
+        elif hasattr(item, "name"):
+            self.fname = item.name
+        else:
+            self.fname = '<unknown file name>'
+        # Actual message
+        self.msg = self._codeinfo.get(code, 'Unknown result')
+        return
+
+    def __str__(self):
+        if len(self.errinfo) > 0:
+            b = " "
+        else:
+            b = ""
+        return "%s: %s%s'%s'" % (self.msg, self.errinfo, b, self.fname)
